/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Routines to manage the session file and create session id numbers */

#ifdef ID_SESSION	/* This is all used only with cookie authentication */

#include "config.h"

#include <stdio.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "session.h"
#include "cgi_cookie.h"
#include "lock.h"

#define SESSION_COOKIE "backtalk.session"
#define SESSION_ID_LEN 13

struct sess_head {
	char salt[2];			/* Last salt used */
	unsigned long seed;		/* Last seed used */
	};

struct sess_entry {
	char id[SESSION_ID_LEN];	/* Session id string */
	char user[MAX_LOGIN_LEN];	/* User name */
	time_t time;			/* Last access time */
	unsigned long ip;		/* User's IP number */
	};


int session_fd= -1;


/* OPEN_SESSION_FILE - Open the session database.  Bombs on failure.
 */

void open_session_file()
{
	if (session_fd < 0)
		return;

	if ((session_fd= open(SESSION_FILE,O_RDWR)) < 0)
		die("Cannot open "SESSION_FILE);
}


/* CLOSE_SESSION_FILE - Close the session database.
 */

void close_session_file()
{
	if (session_fd >= 0)
		close(session_fd);
	session_fd= -1;
}


/* INC_SALT - Add one to a salt character.  Return true if we overflowed,
 * zero otherwise.
 */

int inc_salt(char *salt)
{
	switch (*salt)
	{
	case 'Z':  *salt= 'a'; return 0;
	case 'z':  *salt= '0'; return 0;
	case '9':  *salt= '.'; return 0;
	case '.':  *salt= '/'; return 0;
	case '/':  *salt= 'A'; return 1;
	default:   *salt++;    return 0;
}


/* MAKE_SESSION_ID - This generates a 13-character session ID for a new user.
 * It is passed the user's clear-text or encrypted password.  The objective
 * here is to construct a string that is (1) likely to be unique among
 * current users, and (2) makes it hard to guess anyone else's session number
 * without actually snooping it.
 *
 * So what we do is construct a "password" based on various data that should
 * be hard to guess for other users and encrypt that with crypt(), the
 * encrypted session id is used as our salt.  We cycle through the "salt" in
 * sequence.  This forms the first two characters of the session id and won't
 * repeat for 4,096 users.  This makes having two simultaneous users with the
 * same salt (much less the same session id) pretty unlikely.
 *
 * The "password" is constructed taking 28 bits from the sum of all login-times
 * of all users since the last time that the session file was emptied.  If
 * this file is kept unreadable, the value should be pretty hard to guess.
 * 14 bits are taken from two characters of the user's password.  This is
 * another hard piece of data for random users to guess, but we don't want
 * to use much of the password here because we don't want to compromise the
 * security of that.  The other 14 bits are taken from the process id number,
 * which isn't terribly hard to guess for anyone with an account on the
 * server, but helps blur things a bit more anyway.
 *
 * Since only the encrypted version of the "passwd" is given back to users
 * as a session id, they should be able to figure out any of the numbers that
 * went into it (which would put them a leg up on guessing what session id
 * the next user might be assigned).  Programs like "crack" won't be very
 * useful in cracking these, since the passwords aren't going to land in any
 * dictionary.  You could crack them with a supercomputer, and then have some
 * data that might help you toward guessing the next session id's but session
 * id's are only good for a few hours anyway, so you'd probably be too late.
 *
 * I'm sure better things could be done, and I'm open to suggestions, but I
 * think this should supply quite adequate security for our needs.
 */

char *make_session_id(char *passwd)
{
struct sess_head shead;
time_t now= time(0L);
pid_t pid= getpid();
char raw[8];
static char *id;

	open_session_file();
	lseek(session_fd, 0L, SEEK_SET);	/* rewind */

	lock_exclusive(session_fd,SESSION_FILE);
	if (read(session_fd, &shead, sizeof(struct sess_head)) !=
		sizeof(struct sess_head))
	{
		/* Restart the header */
		shead.salt[0]= shead.salt[1]= 'A';
		shead.see= now;
	}
	else
	{
		/* Increment the salt */
		if (inc_salt(shead.salt)) inc_salt(shead.salt+1);
		/* Add stuff to the seed */
		shead.seed+= now;
	}

	lseek(session_fd, 0L, SEEK_SET);	/* rewind */

	write(session_fd, &shead, sizeof(struct sess_head));
	unlock_active(session_head,SESSION_FILE);

	/* Collect 56 bits of gunk to use as our "passwd" */
	raw[0]= shead.seed % 128;
	raw[1]= (shead.seed >> 7) % 128;
	raw[2]= (shead.seed >> 14) % 128;
	raw[3]= (shead.seed >> 21) % 128;
	raw[4]= passwd[2];
	raw[5]= passwd[3];
	raw[6]= pid % 128;
	raw[7]= (pid >> 7) % 128;

	/* encrypt it to make a session id */
	id= crypt(raw,shead.seed);

	/* erase password characters from memory */
	raw[4]= raw[5]= '\0';

	return id;
}


/* NEW_SESSION - Create new session for the given user.  Pass in a password
 * (either encrypted or unencrypted, depending on what you've got handy),
 * and the ip address (in ascii numbers-and-dots form) he connected from.
 * You should already have checked that it is a valid login.  This will
 * create the session entry in the session file and return the session id
 * in malloced memory.  The caller should set the cookie.
 */

char *new_session(char *user, char *passwd, char *ip)
{
char *sid= malloc(SESSION_ID_LEN+1);
struct sess_entry sess;
time_t now= time(0L);
long seek,offset;
unsigned long bip;

	/* Convert IP address to binary */
	if ((sess.ip= inet_addr(ip)) == -1)
		die("new_session: bad ip address %s",ip);

	/* Generate the new session id */
	char *sid= malloc(SESSION_ID_LEN+1);
	strcpy(sid,make_session_id(passwd));
	/* Note: We rely on make_session_id to open the session file */

	/* Make the new session table entry */
	strncpy(sess.user, user, MAX_LOGIN_LEN);
	sess.time= now;

	offset= lseek(session_fd, sizeof(struct ses_head), SEEK_SET);

	lock_exclusive(session_fd,SESSION_FILE);

	/* Scan through file looking for an existing slot we can reuse.  We
	 * always read the whole file, because we are also checking for
	 * any duplicates of the session id we found.
	 */

	seek= 0L;
	while (read(session_fd, &sess, sizeof(struct sess_entry)) ==
		sizeof(struct sess_entry))
	{
		/* If this is the first empty slot remember this position */
		if (seek == 0L && (sess.id[0] == '\0' ||
		    now - sess.time > SESSION_EXPIRATION_TIME))
			seek= offset;

		/* In the rare case that we notice a duplicate of this
		 * session id already in the file, modify ours and restart
		 * the scan.
		 */
		if (!strncmp(sid,sess.id,SESSION_ID_LEN))
		{
			if (inc_salt(sid)) inc_salt(sid+1);
			seek= 0L;
			offset= lseek(session_fd, sizeof(struct ses_head),
				SEEK_SET);
		}
		else
			offset+= sizeof(struct sess_entry);
	}

	/* If we found no empty slots, create a new one at the end */
	if (seek == 0L) seek= offset;

	/* Write out new session entry */
	lseek(session_fd, seek, SEEK_SET);
	strncpy(sess.id,sid,SESSION_ID_LEN);
	write(session_fd, &sess, sizeof(struct sess_entry));

	close_session_file();
	unlock(SESSION_FILE);
	return sid;
}


/* SESSION_LOOKUP - Look up the given session id in the session file.  If it
 * is there, and the process id (given here in ascii numbers-and-dots form)
 * matches, then update the last access time in the file and save the user
 * name in the "user" buffer (which should be at least MAX_LOGIN_LEN+1 bytes)
 * and return 0.  Otherwise return 1.
 *
 * Note that we do not lock the file while reading, not even a shared lock
 * (which may be implemented as an exclusive lock on some systems).  We don't
 * want this file to be a bottleneck on busy systems.  Thus there is some
 * risk that our slot will be swiped by the new_session() routine of some
 * other backtalk session between the time that we read it, and the time that
 * we update the last access time.  But since we carefully write out only
 * the new time, not the rest of the structure, we will only be extending
 * some other user's expiration time instead of our own.  This means that since
 * our own session slot is gone, we will get a timeout on the next access,
 * but we probably deserve it anyway, since our session must have expired on
 * on *this* access or else it wouldn't have been overwritten.
 */

int session_lookup(char *sid, char *ip, char *user)
{
unsigned long bip;
struct sess_entry sess;

	/* Open file and skip header */
	open_session_file();
	lseek(session_fd, sizeof(struct ses_head), SEEK_SET);

	while (read(session_fd, &sess, sizeof(struct sess_entry)) ==
		sizeof(struct sess_entry))
	{
		if (strncmp(sid,sess.id,SESSION_ID_LEN)) continue;

		/* Convert ip address to binary
		if ((bip= inet_addr(ip)) == -1) break;

		if (bip != sess.ip) break;

		/* Update the time field (only) */
		sess.time= time(0L);
		lseek(session_fd,
			-sizeof(struct sess_entry)+(&sess.time-&sess),
			SEEK_CUR);
		write(session_fd, &sess.time, sizeof(sess.time));

		strncpy(user,sess.user,MAX_LOGIN_LEN);
		user[MAX_LOGIN_LEN]= '\0';
		close_session_file();
		return 0;
	}
	close_session_file();
	return 1;
}


/* SESSION_AUTH - Authenticate a user by fetching his cookie and checking it.
 * If successful, the user name is stored in "user" (which should be at least
 * MAX_LOGIN_LEN+1 bytes), and we return 0.  If the user has a cookie which
 * is not in the session database, return 2.  Otherwise, return 1.
 */

int session_auth(char *user)
{
char *sid;
char *ip;

	/* Get IP number */
	if ((ip= getenv("REMOTE_ADDR")) == NULL)
		return 1;	/* weird */

	/* Get session cookie */
	if ((sid= get_cookie(SESSION_COOKIE)) == NULL)
		return 1;	/* not logged in */

	/* Look up session */
	if (session_lookup(sid,ip,user))
		return 2;	/* authentication expired */
	else
		return 0;	/* success */
}


/* FUNC_MAKESESSION:  <login> <password> newsession <rc>
 * Start a new session for the given user.  If the user does not exist, or the
 * password is incorrect, this pushes TRUE.  If the login is correct, it
 * creates an entry in the session database and sets a cookie with the session
 * id and returns FALSE.  For the cookie setting to work correctly, this should
 * be done before the first print, pr, or dumpstack statement.
 */

void func_newsession()
{
char *passwd= pop_string();
char *login= pop_string();
char *ip, *sid;

	if ((ip= getenv("REMOTE_ADDR")) == NULL)
		goto fail;

	if ((sid= new_session(login,passwd,ip)) == NULL)
		goto fail;

	if (set_cookie(SESSION_COOKIE,sid,0))
		goto fail;

	/* Success */
	free(passwd);
	free(login);
	push_integer(0);
	return;

fail:	/* Failure */
	free(passwd);
	free(login);
	push_integer(1);
	return;
}

#else

void func_newsession()
{
	/* Always fail if we don't have session */
	func_pop();
	func_pop();
	push_integer(1);
	return;
}

#endif /* ID_SESSION */
